from django.contrib.auth.forms import AuthenticationForm as ClassicSigninForm

from .base import SigninFlow


class ClassicSignin(SigninFlow):
    """Classic login flow borrowed from Django built-in AuthenticationForm."""

    form = ClassicSigninForm



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class SitegateConfig(AppConfig):
    """Sitegate configuration."""

    name = 'sitegate'
    verbose_name = _('Sign up & Sign in')



from uuid import uuid4

from django.db import models, IntegrityError
from django.conf import settings
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import python_2_unicode_compatible

from etc.models import InheritedModel


USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


@python_2_unicode_compatible
class BlacklistedDomain(models.Model):

    domain = models.CharField(_('Domain name'), max_length=253, unique=True)
    enabled = models.BooleanField(
        _('Enabled'),
        help_text=_('If enabled visitors won\'t be able to sign up with this domain name in e-mail.'),
        db_index=True, default=True)

    class Meta(object):
        verbose_name = _('Blacklisted domain')
        verbose_name_plural = _('Blacklisted domains')

    @classmethod
    def is_blacklisted(cls, email):
        domain = email.split('@', 1)[1].lower()

        # 'some.denied.co.uk' -> ['some.denied.co.uk', 'denied.co.uk', 'co.uk']
        sub_domains = [domain]
        for i in range(domain.count('.') - 1):
            sub_domains.append(sub_domains[i].split('.', 1)[-1])

        return cls.objects.filter(enabled=True, domain__in=sub_domains).exists()

    def __str__(self):
        return self.domain


@python_2_unicode_compatible
class ModelWithCode(models.Model):

    code = models.CharField('dummy', max_length=128, unique=True, editable=False)
    time_created = models.DateTimeField(_('Date created'), auto_now_add=True)
    time_accepted = models.DateTimeField(_('Date accepted'), null=True, editable=False)
    expired = models.BooleanField(_('Expired'), help_text='dummy', db_index=True, default=False)

    class Meta(object):
        abstract = True

    def __str__(self):
        return self.code

    def save(self, force_insert=False, force_update=False, **kwargs):
        if self.code == '':
            while True:
                self.code = self.generate_code()
                try:
                    super(ModelWithCode, self).save(force_insert, force_update, **kwargs)
                except IntegrityError:
                    pass
                else:
                    break
        else:
            super(ModelWithCode, self).save(force_insert, force_update, **kwargs)

    @classmethod
    def is_valid(cls, code):
        try:
            return cls.objects.get(code=code, expired=False)
        except (cls.MultipleObjectsReturned, cls.DoesNotExist):
            return False

    @staticmethod
    def generate_code():
        return str(uuid4()).replace('-', '')


class InvitationCode(InheritedModel, ModelWithCode):

    creator = models.ForeignKey(
        USER_MODEL, related_name='creators', verbose_name=_('Creator'), on_delete=models.CASCADE)
    acceptor = models.ForeignKey(
        USER_MODEL, related_name='acceptors', verbose_name=_('Acceptor'), null=True, blank=True, editable=False,
        on_delete=models.CASCADE)

    class Meta(object):
        verbose_name = _('Invitation code')
        verbose_name_plural = _('Invitation codes')

    class Fields(object):
        code = _('Invitation code')
        expired = {'help_text': _('Visitors won\'t be able to sign up with an expired code.')}

    @classmethod
    def add(cls, creator):
        new_code = cls(creator=creator)
        new_code.save(force_insert=True)
        return new_code

    @classmethod
    def accept(cls, code, acceptor):
        return cls.objects.filter(code=code).update(acceptor=acceptor, expired=True, time_accepted=timezone.now())


class EmailConfirmation(InheritedModel, ModelWithCode):

    user = models.ForeignKey(USER_MODEL, verbose_name=_('User'), on_delete=models.CASCADE)

    class Meta(object):
        verbose_name = _('Activation code')
        verbose_name_plural = _('Activation codes')

    class Fields(object):
        code = _('Activation code')
        expired = {'help_text': _('Expired codes couldn\'t be used for repeated account activations.')}

    @classmethod
    def add(cls, user):
        new_code = cls(user=user)
        new_code.save(force_insert=True)
        return new_code

    def activate(self):
        self.expired = True
        self.time_accepted = timezone.now()
        self.save()

        user = self.user
        user.is_active = True
        user.save()



from django.shortcuts import redirect
from django.contrib.auth import login

from ..flows_base import FlowsBase
from ..settings import SIGNIN_ENABLED, SIGNIN_DISABLED_TEXT


class SigninFlow(FlowsBase):
    """Base class for sign in flows."""

    flow_type = 'signin'
    enabled = SIGNIN_ENABLED
    disabled_text = SIGNIN_DISABLED_TEXT

    def handle_form_valid(self, request, form):
        login(request, form.get_user())
        redirect_to = self.flow_args.get('redirect_to', self.default_redirect_to)
        if redirect_to:  # TODO Handle lambda variant with user as arg.
            return redirect(redirect_to)



from collections import OrderedDict

from django import forms
from django.contrib.auth import authenticate, login
from etc.toolbox import set_form_widgets_attrs


class FlowsBase(object):
    """Base class for signup and sign in flows."""

    flow_type = None
    default_form_template = 'form_as_p'
    default_redirect_to = '/'
    enabled = None
    disabled_text = None

    def __init__(self, **kwargs):
        if not getattr(self, 'form', False):
            raise NotImplementedError('Please define `form` attribute in your `%s` class.' % self.__class__.__name__)
        self.flow_args = kwargs

    def get_template_name(self, template_name):
        """Returns template path, either from a shortcut built-in template
        name (`form_as_p`) or a template path (`my_sitegate/my_tpl.html`).

        Note that template path can include one `%s` placeholder. In that case
        it will be replaced with flow type (`signin` or `signup`).

        """
        if template_name is None:
            # Build default template path.
            template_name = self.default_form_template
        if '.html' not in template_name:  # Shortcut, e.g.: .
            template_name = '%s%s.html' % ('sitegate/%s/', template_name)
        if '%s' in template_name:  # Fill in the flow type placeholder.
            template_name = template_name % self.flow_type
        return template_name

    def handle_form_valid(self, request, form):
        """"""
        raise NotImplementedError(
            'Please implement `handle_form_valid` method in your `%s` class.' % self.__class__.__name__)

    def respond_for(self, view_function, args, kwargs):
        """Returns a response for the given view & args."""

        request = args[0]
        form = self.get_requested_form(request)
        if form.is_valid():
            result = self.handle_form_valid(request, form)
            if result:
                return result

        self.update_request(request, form)
        return view_function(*args, **kwargs)

    def update_request(self, request, form):
        """Updates Request object with flows forms."""
        forms_key = '%s_forms' % self.flow_type

        # Use ordered forms dict in case _formNode wants to fetch the first defined.
        flow_dict = OrderedDict()
        try:
            flow_dict = request.sitegate[forms_key]
        except AttributeError:
            request.sitegate = {}
        except KeyError:
            pass

        flow_dict[self.get_flow_name()] = form
        request.sitegate[forms_key] = flow_dict

    @classmethod
    def get_flow_name(cls):
        """Returns sign up flow identifier from flow class name.
        Example: `classic` for `ClassicSignup`.

        """
        return cls.__name__

    @staticmethod
    def login_generic(request, username, password):
        """Helper method. Generic login with username and password."""
        user = authenticate(username=username, password=password)
        if user is not None and user.is_active:
            login(request, user)
            return True
        return False

    def get_arg_or_attr(self, name, default=None):
        """Returns flow argument, as provided with sitegate decorators
           or attribute set as a flow class attribute or default."""
        if name in self.flow_args:
            return self.flow_args[name]
        try:
            return getattr(self, name)
        except AttributeError:
            return default

    def get_requested_form(self, request):
        """Returns an instance of a form requested."""
        flow_name = self.get_flow_name()
        flow_key = '%s_flow' % self.flow_type
        form_data = None

        if request.method == 'POST' and request.POST.get(flow_key, False) and request.POST[flow_key] == flow_name:
            form_data = request.POST

        form = self.init_form(form_data, widget_attrs=self.flow_args.get('widget_attrs', None),
                              template=self.get_template_name(self.flow_args.get('template', None)))
        # Attach flow identifying field to differentiate among several possible forms.
        form.fields[flow_key] = forms.CharField(required=True, initial=flow_name, widget=forms.HiddenInput)
        form.flow_enabled = self.enabled
        form.flow_disabled_text = self.disabled_text
        return form

    def init_form(self, form_data, widget_attrs=None, template=None):
        """Constructs, populates and returns a form."""
        form = self.form(data=form_data)
        form.template = template
        # Attach flow attribute to have access from flow forms (usually to call get_arg_or_attr())
        form.flow = self
        if widget_attrs is not None:
            set_form_widgets_attrs(form, widget_attrs)
        return form




ANONYMOUS_MAILERS = [
    '3a.by',
    'dispostable.com',
    'grandmamail.com',
    'harakirimail.com',
    'lackmail.ru',
    'maildrop.cc',
    'mailinator.com',
]
"""Known anonymous mailing hosts to be used in BlacklistedDomain."""



from django.shortcuts import redirect

from ..flows_base import FlowsBase
from ..signals import sig_user_signup_success, sig_user_signup_fail
from ..settings import SIGNUP_ENABLED, SIGNUP_DISABLED_TEXT


class SignupFlow(FlowsBase):
    """Base class for signup flows."""

    flow_type = 'signup'
    auto_signin = True
    activate_user = True
    enabled = SIGNUP_ENABLED
    disabled_text = SIGNUP_DISABLED_TEXT

    def handle_form_valid(self, request, form):
        flow_name = self.get_flow_name()
        if not self.get_arg_or_attr('activate_user'):
            form.instance.is_active = False
        signup_result = self.add_user(request, form)
        if signup_result:
            sig_user_signup_success.send(self, signup_result=signup_result, flow=flow_name, request=request)
            if self.get_arg_or_attr('auto_signin'):
                self.sign_in(request, form, signup_result)
            redirect_to = self.flow_args.get('redirect_to', self.default_redirect_to)
            if redirect_to:  # TODO Handle lambda variant with user as arg.
                return redirect(redirect_to)
        else:
            sig_user_signup_fail.send(self, signup_result=signup_result, flow=flow_name, request=request)

    def add_user(self, request, form):
        """Adds (creates) user using form data."""
        raise NotImplementedError('Please implement `add_user()` method in your `%s` class.' % self.__class__.__name__)

    def sign_in(self, request, form, signup_result):
        """Signs in a user."""
        raise NotImplementedError('Please implement `login()` method in your `%s` class.' % self.__class__.__name__)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import etc.models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='BlacklistedDomain',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('domain', models.CharField(unique=True, max_length=253, verbose_name='Domain name')),
                ('enabled', models.BooleanField(default=True, help_text="If enabled visitors won't be able to sign up with this domain name in e-mail.", db_index=True, verbose_name='Enabled')),
            ],
            options={
                'verbose_name': 'Blacklisted domain',
                'verbose_name_plural': 'Blacklisted domains',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='EmailConfirmation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('code', models.CharField(verbose_name='dummy', unique=True, max_length=128, editable=False)),
                ('time_created', models.DateTimeField(auto_now_add=True, verbose_name='Date created')),
                ('time_accepted', models.DateTimeField(verbose_name='Date accepted', null=True, editable=False)),
                ('expired', models.BooleanField(default=False, help_text="Expired codes couldn't be used for repeated account activations.", db_index=True, verbose_name='Expired')),
                ('user', models.ForeignKey(verbose_name='User', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'Activation code',
                'verbose_name_plural': 'Activation codes',
            },
            bases=(etc.models.InheritedModel, models.Model),
        ),
        migrations.CreateModel(
            name='InvitationCode',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('code', models.CharField(verbose_name='dummy', unique=True, max_length=128, editable=False)),
                ('time_created', models.DateTimeField(auto_now_add=True, verbose_name='Date created')),
                ('time_accepted', models.DateTimeField(verbose_name='Date accepted', null=True, editable=False)),
                ('expired', models.BooleanField(default=False, help_text="Visitors won't be able to sign up with an expired code.", db_index=True, verbose_name='Expired')),
                ('acceptor', models.ForeignKey(related_name='acceptors', blank=True, editable=False, to=settings.AUTH_USER_MODEL, null=True, verbose_name='Acceptor', on_delete=models.CASCADE)),
                ('creator', models.ForeignKey(related_name='creators', verbose_name='Creator', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'Invitation code',
                'verbose_name_plural': 'Invitation codes',
            },
            bases=(etc.models.InheritedModel, models.Model),
        ),
    ]



from django import VERSION
from django.conf.urls import url

from .views import verify_email


def get_sitegate_urls():
    """Returns sitegate urlpatterns, that can be attached
    to urlpatterns of a project:

        # Example from urls.py.

        from sitegate.toolbox import get_sitegate_urls

        urlpatterns = patterns('',
            ...
            url(r'^login/$', 'apps.views.login', name='login'),
            ...
        ) + get_sitegate_urls()  # Attach.

    """
    url_verify = url(r'^verify_email/(?P<code>\S+)/$', verify_email, name='verify_email')

    if VERSION >= (1, 9):
        return [url_verify]

    from django.conf.urls import patterns
    return patterns('', url_verify)



from django.contrib import admin

from .models import InvitationCode, BlacklistedDomain, EmailConfirmation


class InvitationCodeAdmin(admin.ModelAdmin):
    list_display = ('code', 'time_created', 'time_accepted', 'expired')
    list_display_links = ('code',)
    search_fields = ('code', 'creator', 'acceptor')
    list_filter = ('time_accepted',)
    ordering = ('-time_created',)
    date_hierarchy = 'time_created'
admin.site.register(InvitationCode, InvitationCodeAdmin)


class EmailConfirmationAdmin(admin.ModelAdmin):
    list_display = ('code', 'time_created', 'time_accepted', 'expired')
    list_display_links = ('code',)
    search_fields = ('code', 'user')
    list_filter = ('time_accepted',)
    ordering = ('-time_created',)
    date_hierarchy = 'time_created'
admin.site.register(EmailConfirmation, EmailConfirmationAdmin)


class BlacklistedDomainAdmin(admin.ModelAdmin):
    list_display = ('domain', 'enabled')
    list_display_links = ('domain',)
    search_fields = ('domain',)
    list_filter = ('enabled',)
    ordering = ('domain',)
admin.site.register(BlacklistedDomain, BlacklistedDomainAdmin)



"""This file contains tests for sitegate."""
import unittest
from uuid import uuid4

from django.contrib.auth import get_user_model

try:
    from django.http.response import HttpResponse
except ImportError:
    from django.http import HttpResponse

try:
    from django.contrib.auth.tests.custom_user import CustomUser
except ImportError:
    # Since 1.10
    pass

from django import VERSION as DJANGO_VERSION
from django.conf.urls import url

try:
    from django.urls import reverse, get_urlconf, set_urlconf, URLPattern

except ImportError:  # Django < 2.0
    from django.core.urlresolvers import reverse, RegexURLPattern as URLPattern, get_urlconf, set_urlconf

from django.http import HttpResponseRedirect
from django.template.base import Template
from django.template.context import Context
from django.test import TestCase
from django.test.client import RequestFactory
from django.test.utils import override_settings
from django.utils import timezone

from sitegate.decorators import signin_view, signup_view, redirect_signedin, sitegate_view
from sitegate.signup_flows.classic import *
from sitegate.signup_flows.modern import *
from sitegate.signin_flows.classic import *
from sitegate.signin_flows.modern import *
from sitegate.models import *
from sitegate.toolbox import *
from sitegate.views import verify_email, messages
from sitegate.utils import USER, get_username_field


class MockUser(object):

    def __init__(self, logged_in):
        self.logged_in = logged_in

    def is_authenticated(self):
        return self.logged_in


@signup_view(activate_user=False, auto_signin=False, validate_email_domain=False, redirect_to='ok')
def register(request):
    return response_from_string(request, "{% load sitegate %}{% sitegate_signup_form %}")


@redirect_signedin('fail')
@signin_view(redirect_to='ok')
def login(request):
    return response_from_string(request, "{% load sitegate %}{% sitegate_signin_form %}")


urlpatterns = [
    url(r'entrance/', lambda r: None),
    url(r'^ok/$', lambda r: HttpResponse('ok'), name='ok'),
    url(r'^fail/$', lambda r: HttpResponse('fail'), name='fail'),
    url(r'^login/$', login, name='login'),
    url(r'^register/$', register, name='register'),
] + get_sitegate_urls()


if VERSION < (1, 10):
    from django.conf.urls import patterns
    urlpatterns.insert(0, '')
    urlpatterns = patterns(*urlpatterns)


def response_from_string(request, string):
    return HttpResponse(Template(string).render(Context({'request': request})))


URL_REGISTER = reverse('register')
URL_LOGIN = reverse('login')

MESSAGE_SUCCESS = None
MESSAGE_ERROR = None


def catch_message_success(*args, **kwargs):
    global MESSAGE_SUCCESS
    MESSAGE_SUCCESS = True


def catch_message_error(*args, **kwargs):
    global MESSAGE_ERROR
    MESSAGE_ERROR = True


# Mimic messages contrib.
messages.success = catch_message_success
messages.error = catch_message_error


@override_settings(USE_I18N=False)
class ViewsTest(TestCase):

    def setUp(self):
        # user credentials
        self._username = self._regenerate_email()
        self._password = 'qwerty'

    def _register_user(self, **kwargs):
        default_data = {'email': self._email, 'password1': self._password, "signup_flow": "ModernSignup"}
        default_data.update(kwargs)
        return self.client.post(URL_REGISTER, default_data)

    def _login(self, **kwargs):
        default_data = {'username': self._email, 'password': self._password, "signin_flow": "ModernSignin"}
        default_data.update(kwargs)
        return self.client.post(URL_LOGIN, default_data)

    def _regenerate_email(self, name_len=20):
        self._email = '{}@mail.com'.format('a' * name_len)
        return self._email

    @unittest.skipIf(settings.AUTH_USER_MODEL != 'auth.User', 'Custom user model is not supported for this test')
    def test_modern_signin(self):

        # username is too long
        if DJANGO_VERSION < (1, 5, 0):
            self._regenerate_email(200)
            response = self._login()
            self.assertEqual(response.status_code, 200)
            self.assertFormError(response, 'signin_form', 'username', 'Ensure this value has at most 30 characters'
                                                                      ' (it has %s).' % len(self._email))

        self._regenerate_email()

        # login with wrong credentials
        response = self._login()
        self.assertEqual(response.status_code, 200)

        if DJANGO_VERSION >= (1, 5, 0):  # 1.4 has html markup in actual errors description
            self.assertFormError(response, 'signin_form', None, 'Please enter a correct username and password. '
                                                                'Note that both fields may be case-sensitive.')

        # create deactivated test user
        response = self._register_user()
        self.assertRedirects(response, reverse('ok'))
        response = self._login()

        content = response.content.decode()
        if 'This account is inactive.' not in content and 'Please enter a correct username and password' not in content:
            self.assertTrue(False)

        # activate user and try to login again
        get_user_model().objects.filter(username=self._username).update(is_active=True)
        response = self._login()
        self.assertRedirects(response, reverse('ok'))

        # how about to login when user is already signed in?
        response = self._login()
        self.assertRedirects(response, reverse('fail'))
        self.client.logout()

        # more than one user with this e-mail
        get_user_model().objects.create(username='dummy', email=self._email)
        response = self._login()
        self.assertFormError(response, 'signin_form', None, 'There is more than one user with this e-mail. '
                                                            'Please use your username to log in.')

    @unittest.skipIf(settings.AUTH_USER_MODEL != 'auth.User', 'Custom user model is not supported for this test')
    def test_modern_signup(self):
        # create deactivated user
        response = self._register_user()
        self.assertRedirects(response, reverse('ok'))

        # auto_signin = False
        self.assertFalse(str(self.client.cookies).strip(), 'Cookies not empty')

        # he can't login yet
        login_result = self.client.login(username=self._username, password=self._password)
        self.assertFalse(login_result)

        # check fields
        user = get_user_model()._default_manager.get(username=self._username)
        self.assertEqual(user.email, self._email)
        self.assertEqual(user.username, self._username)

        # try to sign up with the same data for second time
        response = self._register_user()
        self.assertEqual(response.status_code, 200)
        self.assertFormError(response, 'signup_form', 'email', 'A user with that e-mail already exists.')

        # try to signup with email too long
        self._email = '%s@some.com' % ('long' * 100)
        response = self._register_user()
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'E-mail length should be no more than')


class DecoratorsTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # set urlconf to test's one
        cls.urlconf_bak = get_urlconf()
        set_urlconf('sitegate.tests')

    @classmethod
    def tearDownClass(cls):
        set_urlconf(cls.urlconf_bak)

    def setUp(self):
        self.req_factory = RequestFactory()

    def test_signin(self):
        # simple decoration
        response = signin_view(lambda req: req)(self.req_factory.get('/signin/'))
        self.assertTrue(hasattr(response, 'sitegate'))
        self.assertTrue('signin_forms' in response.sitegate)
        self.assertFalse('signup_forms' in response.sitegate)

        # simple decoration + view with args
        response = signin_view(lambda req, a1, a2='b': (req, a1, a2))(self.req_factory.get('/signin/'), 10, a2=20)
        self.assertTrue(hasattr(response[0], 'sitegate'))
        self.assertTrue('signin_forms' in response[0].sitegate)
        self.assertFalse('signup_forms' in response[0].sitegate)
        self.assertEqual(response[1], 10)
        self.assertEqual(response[2], 20)

        # stacking
        signin_view1 = signin_view(flow=ClassicSignin)
        signin_view2 = signin_view(flow=ModernSignin)
        response = signin_view1(signin_view2(lambda req: req))(self.req_factory.get('/signin/'))
        self.assertTrue(hasattr(response, 'sitegate'))
        self.assertTrue('signin_forms' in response.sitegate)
        self.assertFalse('signup_forms' in response.sitegate)
        self.assertTrue(len(response.sitegate['signin_forms']) == 2)
        dict_keys = list(response.sitegate['signin_forms'].keys())
        self.assertIsInstance(response.sitegate['signin_forms'][dict_keys[0]], ClassicSignin.form)
        self.assertIsInstance(response.sitegate['signin_forms'][dict_keys[1]], ModernSignin.form)

        # stacking with args
        signin_view1 = signin_view(flow=ClassicSignin, widget_attrs={'attr_name1': 'attr_value1'})
        signin_view2 = signin_view(flow=ModernSignin, widget_attrs={'attr_name2': 'attr_value2'})
        response = signin_view1(signin_view2(lambda req: req))(self.req_factory.get('/signin/'))
        self.assertTrue(hasattr(response, 'sitegate'))
        field = response.sitegate['signin_forms']['ClassicSignin'].fields.get('password')
        self.assertIn('attr_name1', field.widget.attrs)
        field = response.sitegate['signin_forms']['ModernSignin'].fields.get('password')
        self.assertIn('attr_name2', field.widget.attrs)

    def test_signup(self):
        # simple decoration
        response = signup_view(lambda req: req)(self.req_factory.get('/signup/'))
        self.assertTrue(hasattr(response, 'sitegate'))
        self.assertTrue('signup_forms' in response.sitegate)
        self.assertFalse('signin_forms' in response.sitegate)

        # simple decoration + view with args
        response = signup_view(lambda req, a1, a2='b': (req, a1, a2))(self.req_factory.get('/signup/'), 10, a2=20)
        self.assertTrue(hasattr(response[0], 'sitegate'))
        self.assertTrue('signup_forms' in response[0].sitegate)
        self.assertFalse('signin_forms' in response[0].sitegate)
        self.assertEqual(response[1], 10)
        self.assertEqual(response[2], 20)

        # stacking
        signup_view1 = signup_view(flow=ClassicSignup)
        signup_view2 = signup_view(flow=ModernSignup)
        response = signup_view1(signup_view2(lambda req: req))(self.req_factory.get('/signup/'))
        self.assertTrue(hasattr(response, 'sitegate'))
        self.assertTrue('signup_forms' in response.sitegate)
        self.assertFalse('signin_forms' in response.sitegate)
        self.assertTrue(len(response.sitegate['signup_forms']) == 2)
        dict_keys = list(response.sitegate['signup_forms'].keys())
        self.assertIsInstance(response.sitegate['signup_forms'][dict_keys[0]], ClassicSignup.form)
        self.assertIsInstance(response.sitegate['signup_forms'][dict_keys[1]], ModernSignup.form)

        # stacking with args
        signup_view1 = signup_view(flow=ClassicSignup, widget_attrs={'attr_name1': 'attr_value1'})
        signup_view2 = signup_view(flow=ModernSignup, widget_attrs={'attr_name2': 'attr_value2'})
        response = signup_view1(signup_view2(lambda req: req))(self.req_factory.get('/signup/'))
        self.assertTrue(hasattr(response, 'sitegate'))
        field = response.sitegate['signup_forms']['ClassicSignup'].fields.get('password1')
        self.assertIn('attr_name1', field.widget.attrs)
        field = response.sitegate['signup_forms']['ModernSignup'].fields.get('password1')
        self.assertIn('attr_name2', field.widget.attrs)

    def test_signinup(self):
        # simple decoration
        response = signin_view(signup_view(lambda req: req))(self.req_factory.get('/entrance/'))
        self.assertTrue(hasattr(response, 'sitegate'))
        self.assertTrue('signup_forms' in response.sitegate)
        self.assertTrue('signin_forms' in response.sitegate)

        # simple decoration + view with args
        response = signin_view(signup_view(lambda req, a1, a2='b': (req, a1, a2)))(self.req_factory.get('/entrance/'), 10, a2=20)
        self.assertTrue(hasattr(response[0], 'sitegate'))
        self.assertTrue('signup_forms' in response[0].sitegate)
        self.assertTrue('signin_forms' in response[0].sitegate)
        self.assertEqual(response[1], 10)
        self.assertEqual(response[2], 20)

        # stacking with args
        view_signin = signin_view(widget_attrs={'attr_name1': 'attr_value1'})
        view_signup = signup_view(widget_attrs={'attr_name2': 'attr_value2'})
        response = view_signin(view_signup(lambda req, a1, a2='b': (req, a1, a2)))(self.req_factory.get('/entrance/'), 10, a2=20)
        self.assertTrue(hasattr(response[0], 'sitegate'))
        self.assertTrue('signup_forms' in response[0].sitegate)
        self.assertTrue('signin_forms' in response[0].sitegate)
        self.assertEqual(response[1], 10)
        self.assertEqual(response[2], 20)
        field = response[0].sitegate['signin_forms']['ModernSignin'].fields.get('password')
        self.assertIn('attr_name1', field.widget.attrs)
        field = response[0].sitegate['signup_forms']['ModernSignup'].fields.get('password1')
        self.assertIn('attr_name2', field.widget.attrs)

    def test_redirect(self):
        # logged in
        # default redirect
        request = self.req_factory.get('/entrance/')
        request.user = MockUser(logged_in=True)
        response = redirect_signedin(lambda req: req)(request)
        self.assertIsInstance(response, HttpResponseRedirect)

        # not logged in
        request = self.req_factory.get('/entrance/')
        request.user = MockUser(logged_in=False)
        response = redirect_signedin(lambda req: req)(request)
        self.assertNotIsInstance(response, HttpResponseRedirect)

        # paametrized decoration
        request = self.req_factory.get('/entrance/')
        request.user = MockUser(logged_in=True)
        redirector = redirect_signedin('/abc/')
        response = redirector(lambda req: req)(request)
        self.assertIsInstance(response, HttpResponseRedirect)
        self.assertEqual(response['Location'], '/abc/')

    def test_sitegate(self):
        # logged in
        request = self.req_factory.get('/entrance/')
        request.user = MockUser(True)
        response = sitegate_view(lambda req: req)(request)
        self.assertIsInstance(response, HttpResponseRedirect)

        # not logged in

        # simple decoration
        request = self.req_factory.get('/entrance/')
        request.user = MockUser(False)
        response = sitegate_view(lambda req: req)(request)
        self.assertNotIsInstance(response, HttpResponseRedirect)
        self.assertTrue(hasattr(response, 'sitegate'))
        self.assertTrue('signup_forms' in response.sitegate)
        self.assertTrue('signin_forms' in response.sitegate)

        # simple decoration + view with args
        request = self.req_factory.get('/entrance/')
        request.user = MockUser(False)
        response = sitegate_view(lambda req, a1, a2='b': (req, a1, a2))(request, 10, a2=20)
        self.assertNotIsInstance(response, HttpResponseRedirect)
        self.assertTrue(hasattr(response[0], 'sitegate'))
        self.assertTrue('signup_forms' in response[0].sitegate)
        self.assertTrue('signin_forms' in response[0].sitegate)
        self.assertEqual(response[1], 10)
        self.assertEqual(response[2], 20)

        # parametrized decoration
        request = self.req_factory.get('/entrance/')
        request.user = MockUser(False)
        view_sitegate = sitegate_view(widget_attrs={'attr_name1': 'attr_value1'})
        response = view_sitegate(lambda req, a1, a2='b': (req, a1, a2))(request, 10, a2=20)
        self.assertTrue(hasattr(response[0], 'sitegate'))
        field = response[0].sitegate['signin_forms']['ModernSignin'].fields.get('password')
        self.assertIn('attr_name1', field.widget.attrs)
        field = response[0].sitegate['signup_forms']['ModernSignup'].fields.get('password1')
        self.assertIn('attr_name1', field.widget.attrs)
        self.assertEqual(response[1], 10)
        self.assertEqual(response[2], 20)


class ClassicWithEmailSignupTest(TestCase):

    @unittest.skipIf(settings.AUTH_USER_MODEL != 'auth.User', 'Custom user model is not supported for this test')
    def test_basic(self):
        flow = ClassicWithEmailSignup()
        self.assertTrue(flow.validate_email_domain)
        self.assertFalse(flow.verify_email)
        self.assertFalse(hasattr(flow, 'schedule_email'))

        flow = ClassicWithEmailSignup(verify_email=True)
        self.assertTrue(flow.flow_args['verify_email'])
        self.assertFalse(flow.flow_args['activate_user'])
        self.assertFalse(flow.flow_args['auto_signin'])
        self.assertTrue(hasattr(flow, 'schedule_email'))

        form = ClassicWithEmailSignupForm({
            'username': 'abcom',
            'email': 'a@b.com',
            'password1': 'qwerty',
            'password2': 'qwerty',
        })
        form.flow = flow

        def fake_schedule_email(text, user, title):
            self.assertEqual(user.username, 'abcom')
            self.assertEqual(title, 'Account activation')

        flow.schedule_email = fake_schedule_email

        new_user = flow.add_user(RequestFactory().get('/'), form)
        self.assertEqual(new_user.username, 'abcom')


class ClassicSignupFormsTest(unittest.TestCase):

    def test_classic_signup_attrs(self):
        f = ClassicSignupForm()
        fields = set(f.fields.keys())

        self.assertTrue(get_username_field() in fields)
        self.assertTrue('password1' in fields)
        self.assertTrue('password2' in fields)

    @unittest.skipIf(settings.AUTH_USER_MODEL != 'auth.User', 'Custom user model is not supported for this test')
    def test_classic_signup_validation(self):
        get_user_model()._default_manager.create(username='test_duplicate', email='test_duplicate@mail.some')

        f = ClassicSignupForm(data={
            'username': 'test_duplicate',
            'email': 'test_duplicate@mail.some',
        })
        f.full_clean()
        self.assertIn('username', f.errors)
        self.assertEqual(len(f.errors), 3)

        f = ClassicSignupForm(data={
            'username': 'test_duplicate',
            'email': 'test_duplicate@mail.some',
            'password1': 'password',
            'password2': 'password',
        })
        f.full_clean()
        self.assertIn('username', f.errors)
        self.assertEqual(len(f.errors), 1)

    def test_simple_classic_signup_attrs(self):
        f = SimpleClassicSignupForm()
        fields = set(f.fields.keys())

        self.assertTrue(get_username_field() in fields)
        self.assertTrue('password1' in fields)
        self.assertFalse('password2' in fields)

    def test_classic_with_email_signup_attrs(self):
        f = ClassicWithEmailSignupForm()
        fields = set(f.fields.keys())

        self.assertTrue(get_username_field() in fields)
        self.assertTrue('email' in fields)
        self.assertTrue('password1' in fields)
        self.assertTrue('password2' in fields)

    def test_simple_classic_with_email_signup_attrs(self):
        f = SimpleClassicWithEmailSignupForm()
        fields = set(f.fields.keys())

        self.assertTrue(get_username_field() in fields)
        self.assertTrue('email' in fields)
        self.assertTrue('password1' in fields)
        self.assertFalse('password2' in fields)


@unittest.skipIf(settings.AUTH_USER_MODEL == 'auth.User', 'Custom user model required')
class ClassicSignupCustomUserFormsTest(TestCase):

    def test_classic_signup_attrs(self):
        f = ClassicSignupForm()
        fields = set(f.fields.keys())

        self.assertTrue('email' in fields)
        self.assertTrue('password1' in fields)
        self.assertTrue('password2' in fields)

    def test_classic_signup_validation(self):
        get_user_model()._default_manager.create(
            email='test_duplicate@mail.some',
            date_of_birth=timezone.now().date(),
        )

        f = ClassicSignupForm(data={
            'email': 'test_duplicate@mail.some',
        })
        f.full_clean()
        self.assertIn('email', f.errors)
        self.assertEqual(len(f.errors), 3)

        f = ClassicSignupForm(data={
            'email': 'test_duplicate@mail.some',
            'password1': 'password',
            'password2': 'password',
        })
        f.full_clean()
        self.assertIn('email', f.errors)
        self.assertEqual(len(f.errors), 1)

    def test_simple_classic_signup_attrs(self):
        f = SimpleClassicSignupForm()
        fields = set(f.fields.keys())

        self.assertTrue('username' not in fields)
        self.assertTrue('email' in fields)
        self.assertTrue('password1' in fields)
        self.assertFalse('password2' in fields)

    def test_classic_with_email_signup_attrs(self):
        f = ClassicWithEmailSignupForm()
        fields = set(f.fields.keys())

        self.assertTrue('username' not in fields)
        self.assertTrue('email' in fields)
        self.assertTrue('password1' in fields)
        self.assertTrue('password2' in fields)

    def test_simple_classic_with_email_signup_attrs(self):
        f = SimpleClassicWithEmailSignupForm()
        fields = set(f.fields.keys())

        self.assertTrue('username' not in fields)
        self.assertTrue('email' in fields)
        self.assertTrue('password1' in fields)
        self.assertFalse('password2' in fields)



class ModernSignupFormsTest(unittest.TestCase):

    def test_modern_signup_attrs(self):
        f = ModernSignupForm()
        fields = set(f.fields.keys())

        self.assertFalse('username' in fields)
        self.assertTrue('email' in fields)
        self.assertTrue('password1' in fields)
        self.assertFalse('password2' in fields)

    def test_invitation_signup_attrs(self):
        f = InvitationSignupForm()
        fields = set(f.fields.keys())

        self.assertFalse('username' in fields)
        self.assertTrue('code' in fields)
        self.assertTrue('email' in fields)
        self.assertTrue('password1' in fields)
        self.assertFalse('password2' in fields)


class GenericSignupFlowTest(unittest.TestCase):

    def test_init_exception(self):
        self.assertRaises(NotImplementedError, SignupFlow)

    def test_getflow_name(self):
        self.assertEqual(ModernSignup.get_flow_name(), 'ModernSignup')


class ClassicSigninFormsTest(unittest.TestCase):

    def test_classic_signin_attrs(self):
        f = ClassicSigninForm()
        fields = set(f.fields.keys())

        self.assertTrue('username' in fields)
        self.assertTrue('password' in fields)


class ModernSigninFormsTest(unittest.TestCase):

    def test_modern_signin_attrs(self):
        f = ModernSigninForm()
        fields = set(f.fields.keys())

        self.assertTrue('username' in fields)
        self.assertTrue('password' in fields)


@unittest.skipIf(settings.AUTH_USER_MODEL != 'auth.User', 'Custom user model is not supported for this test')
class InvitationCodeModelTest(unittest.TestCase):

    def setUp(self):
        self.user = USER(username=str(uuid4()))
        self.user.save()

    def tearDown(self):
        self.user.delete()

    def test_generate_code(self):
        code = InvitationCode.generate_code()
        self.assertIsNotNone(code)

    def test_add(self):
        code = InvitationCode.add(self.user)
        self.assertIsNotNone(code)
        self.assertEqual(code.creator, self.user)
        self.assertIsNone(code.acceptor)
        self.assertIsNotNone(code.time_created)
        self.assertIsNone(code.time_accepted)
        self.assertFalse(code.expired)

    def test_accept(self):
        code = InvitationCode.add(self.user)
        rows = InvitationCode.accept(code.code, self.user)
        self.assertEqual(rows, 1)
        updated_code = InvitationCode.objects.get(code=code.code)
        self.assertEqual(updated_code.acceptor, self.user)
        self.assertIsNotNone(updated_code.time_accepted)
        self.assertTrue(updated_code.expired)


class BlacklistedDomainModelTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        BlacklistedDomain.objects.bulk_create([
            BlacklistedDomain(domain='denied1.com', enabled=False),
            BlacklistedDomain(domain='denied2.com', enabled=True),
            BlacklistedDomain(domain='denied3.com', enabled=True),
            BlacklistedDomain(domain='denied4.co.uk', enabled=True)
        ])

    @classmethod
    def tearDownClass(cls):
        BlacklistedDomain.objects.all().delete()

    def test_is_blacklisted(self):
        self.assertFalse(BlacklistedDomain.is_blacklisted('example1@denied1.com'))
        self.assertTrue(BlacklistedDomain.is_blacklisted('example2@denied2.com'))
        self.assertTrue(BlacklistedDomain.is_blacklisted('example3@some.denied3.com'))
        self.assertTrue(BlacklistedDomain.is_blacklisted('example3@some.denied3.COM'))
        self.assertFalse(BlacklistedDomain.is_blacklisted('example4@co.uk'))
        self.assertTrue(BlacklistedDomain.is_blacklisted('example4@denied4.co.UK'))
        self.assertTrue(BlacklistedDomain.is_blacklisted('example4@some.Denied4.co.UK'))


class ToolboxTest(unittest.TestCase):

    def test_get_sitegate_urls(self):
        urls = get_sitegate_urls()
        self.assertIsInstance(urls, list)
        self.assertTrue(len(urls) == 1)
        self.assertIsInstance(urls[0], URLPattern)


@unittest.skipIf(settings.AUTH_USER_MODEL != 'auth.User', 'Custom user model is not supported for this test')
class ViewsModuleTest(unittest.TestCase):

    def test_verify_email(self):
        request = RequestFactory().get('/')

        code = 42
        result = verify_email(request, code)
        self.assertEqual(result._headers['location'][1], '/')
        self.assertTrue(MESSAGE_ERROR)

        user = USER(username=str(uuid4()))
        user.save()

        code = EmailConfirmation.add(user)
        result = verify_email(request, code, redirect_to='/here/')
        self.assertEqual(result._headers['location'][1], '/here/')
        self.assertTrue(MESSAGE_SUCCESS)



"""This file contains decorators used by sitegate."""
from django.shortcuts import redirect

from .signin_flows.modern import ModernSignin
from .signup_flows.modern import ModernSignup
from .utils import DecoratorBuilder


class FlowBuilder(DecoratorBuilder):

    def __init__(self, flow_cls, args, kwargs):
        self.flow_cls = flow_cls
        super(FlowBuilder, self).__init__(args, kwargs)

    def handle(self, func, args_func, kwargs_func, args_dec, kwargs_dec):
        kwargs_dec_ = dict(kwargs_dec)
        flow_class = kwargs_dec_.pop('flow', None)
        if flow_class is None:
            flow_class = self.flow_cls
        flow_obj = flow_class(**kwargs_dec_)
        return flow_obj.respond_for(func, args_func, kwargs_func)


class RedirectBuilder(DecoratorBuilder):

    def handle(self, func, args_func, kwargs_func, args_dec, kwargs_dec):
        authenticated = args_func[0].user.is_authenticated

        if not isinstance(authenticated, bool):  # Not a property in Django<2.0
            authenticated = authenticated()

        if authenticated:
            args_dec_ = list(args_dec)
            if hasattr(args_dec[0], '__call__'):
                args_dec_.insert(0, '/')
            return redirect(*args_dec_, **kwargs_dec)

        return func(*args_func, **kwargs_func)


def signup_view(*args, **kwargs):
    """Decorator to mark views used for signup."""
    return FlowBuilder(ModernSignup, args, kwargs)


def signin_view(*args, **kwargs):
    """Decorator to mark views used for sign in."""
    return FlowBuilder(ModernSignin, args, kwargs)


def redirect_signedin(*args, **kwargs):
    """Decorator to mark views which should not be accessed by signed in users.
    Example: sign in & sign up pages.

    """
    return RedirectBuilder(args, kwargs)


def sitegate_view(*args_dec, **kwargs_dec):
    """Decorator to mark views used both for signup & sign in."""
    if len(args_dec):  # simple decoration w/o parameters
        return signup_view(signin_view(redirect_signedin(*args_dec, **kwargs_dec)))

    signin = signin_view(**kwargs_dec)
    signup = signup_view(**kwargs_dec)
    return lambda *args, **kwargs: signup(signin(redirect_signedin(*args, **kwargs)))



from django import forms
from django.utils.translation import ugettext_lazy as _

from .base import SigninFlow
from .classic import ClassicSigninForm
from ..utils import USER


class ModernSigninForm(ClassicSigninForm):

    def __init__(self, request=None, *args, **kwargs):
        super(ModernSigninForm, self).__init__(request, *args, **kwargs)
        self.fields['username'].label = u'%s / %s' % (_('Username'), _('Email'))

    def clean(self):
        username = self.cleaned_data.get('username')
        password = self.cleaned_data.get('password')

        if username and password and '@' in username:
            # Let's first try an e-mail auth.
            result = USER._default_manager.filter(email__iexact=username)
            if len(result) == 1:
                self.cleaned_data['username'] = result[0].username
            elif len(result) > 1:
                raise forms.ValidationError(
                    _('There is more than one user with this e-mail. Please use your username to log in.'))

        return super(ModernSigninForm, self).clean()


class ModernSignin(SigninFlow):
    """Modernized login flow based on classic from Django built-in
    with username e-mail authentication support."""

    form = ModernSigninForm





from django.shortcuts import redirect
from django.contrib import messages

from .models import EmailConfirmation
from .settings import SIGNUP_VERIFY_EMAIL_ERROR_TEXT, SIGNUP_VERIFY_EMAIL_SUCCESS_TEXT


def verify_email(request, code, redirect_to=None):
    """Verifies an account activation code a user received by e-mail.

    Requires Messages Django Contrib.

    :param Requset request:
    :param str code:
    :param str redirect_to:
    :return:
    """
    success = False

    valid_code = EmailConfirmation.is_valid(code)
    if valid_code:
        valid_code.activate()
        success = True

    if success:
        messages.success(request, SIGNUP_VERIFY_EMAIL_SUCCESS_TEXT, 'success')
    else:
        messages.error(request, SIGNUP_VERIFY_EMAIL_ERROR_TEXT, 'danger error')

    if redirect_to is None:
        redirect_to = '/'

    return redirect(redirect_to)
